@@ -18,18 +18,20 @@ module Data.List.Extra(
     wordsBy, linesBy,
     breakOn, breakOnEnd, splitOn, split, chunksOf,
     -- * Basics
-    headDef, lastDef, notNull, list, unsnoc, cons, snoc,
-    drop1, dropEnd1, mconcatMap,
+    headDef, lastDef, (!?), notNull, list, unsnoc, cons, snoc,
+    drop1, dropEnd1, mconcatMap, compareLength, comparingLength,
     -- * Enum operations
     enumerate,
     -- * List operations
     groupSort, groupSortOn, groupSortBy,
     nubOrd, nubOrdBy, nubOrdOn,
-    nubOn, groupOn,
+    nubOn, groupOn, groupOnKey,
     nubSort, nubSortBy, nubSortOn,
     maximumOn, minimumOn,
+    sum', product',
+    sumOn', productOn',
     disjoint, disjointOrd, disjointOrdBy, allSame, anySame,
-    repeatedly, firstJust,
+    repeatedly, repeatedlyNE, firstJust,
     concatUnzip, concatUnzip3,
     zipFrom, zipWithFrom, zipWithLongest,
     replace, merge, mergeBy,
@@ -44,12 +46,18 @@ import Data.Tuple.Extra
 import Data.Monoid
 import Numeric
 import Data.Functor
+import Data.Foldable
 import Prelude
+import Data.List.NonEmpty (NonEmpty ((:|)))
 
 
 -- | Apply some operation repeatedly, producing an element of output
 --   and the remainder of the list.
 --
+-- When the empty list is reached it is returned, so the operation
+-- is /never/ applied to the empty input.
+-- That fact is encoded in the type system with 'repeatedlyNE'
+--
 -- > \xs -> repeatedly (splitAt 3) xs  == chunksOf 3 xs
 -- > \xs -> repeatedly word1 (trim xs) == words xs
 -- > \xs -> repeatedly line1 xs == lines xs
@@ -58,6 +66,16 @@ repeatedly f [] = []
 repeatedly f as = b : repeatedly f as'
     where (b, as') = f as
 
+-- | Apply some operation repeatedly, producing an element of output
+--   and the remainder of the list.
+--
+-- Identical to 'repeatedly', but has a more precise type signature.
+repeatedlyNE :: (NonEmpty a -> (b, [a])) -> [a] -> [b]
+repeatedlyNE f [] = []
+repeatedlyNE f (a : as) = b : repeatedlyNE f as'
+    where (b, as') = f (a :| as)
+
+
 
 -- | Are two lists disjoint, with no elements in common.
 --
@@ -69,7 +87,7 @@ disjoint xs = null . intersect xs
 -- | /O((m+n) log m), m <= n/. Are two lists disjoint, with no elements in common.
 --
 -- @disjointOrd@ is more strict than `disjoint`. For example, @disjointOrd@ cannot
--- terminate if both lists are inifite, while `disjoint` can.
+-- terminate if both lists are infinite, while `disjoint` can.
 --
 -- > disjointOrd [1,2,3] [4,5] == True
 -- > disjointOrd [1,2,3] [4,1] == False
@@ -138,6 +156,19 @@ lastDef :: a -> [a] -> a
 lastDef d xs = foldl (\_ x -> x) d xs -- I know this looks weird, but apparently this is the fastest way to do this: https://hackage.haskell.org/package/base-4.12.0.0/docs/src/GHC.List.html#last
 {-# INLINE lastDef #-}
 
+-- | A total variant of the list index function `(!!)`.
+--
+-- > [2,3,4] !? 1    == Just 3
+-- > [2,3,4] !? (-1) == Nothing
+-- > []      !? 0    == Nothing
+(!?) :: [a] -> Int -> Maybe a
+xs !? n
+  | n < 0     = Nothing
+             -- Definition adapted from GHC.List
+  | otherwise = foldr (\x r k -> case k of
+                                   0 -> Just x
+                                   _ -> r (k-1)) (const Nothing) xs n
+{-# INLINABLE (!?) #-}
 
 -- | A composition of 'not' and 'null'.
 --
@@ -196,7 +227,9 @@ enumerate = [minBound..maxBound]
 -- > \i xs -> takeEnd i xs `isSuffixOf` xs
 -- > \i xs -> length (takeEnd i xs) == min (max 0 i) (length xs)
 takeEnd :: Int -> [a] -> [a]
-takeEnd i xs = f xs (drop i xs)
+takeEnd i xs
+    | i <= 0 = []
+    | otherwise = f xs (drop i xs)
     where f (x:xs) (y:ys) = f xs ys
           f xs _ = xs
 
@@ -209,7 +242,9 @@ takeEnd i xs = f xs (drop i xs)
 -- > \i xs -> length (dropEnd i xs) == max 0 (length xs - max 0 i)
 -- > \i -> take 3 (dropEnd 5 [i..]) == take 3 [i..]
 dropEnd :: Int -> [a] -> [a]
-dropEnd i xs = f xs (drop i xs)
+dropEnd i xs
+    | i <= 0 = xs
+    | otherwise = f xs (drop i xs)
     where f (x:xs) (y:ys) = x : f xs ys
           f _ _ = []
 
@@ -222,13 +257,15 @@ dropEnd i xs = f xs (drop i xs)
 -- > \i xs -> uncurry (++) (splitAt i xs) == xs
 -- > \i xs -> splitAtEnd i xs == (dropEnd i xs, takeEnd i xs)
 splitAtEnd :: Int -> [a] -> ([a], [a])
-splitAtEnd i xs = f xs (drop i xs)
+splitAtEnd i xs
+    | i <= 0 = (xs, [])
+    | otherwise = f xs (drop i xs)
     where f (x:xs) (y:ys) = first (x:) $ f xs ys
           f xs _ = ([], xs)
 
 
 -- | 'zip' against an enumeration.
---   Never truncates the output - raises an error if the enumeration runs out.
+--   Truncates the output if the enumeration runs out.
 --
 -- > \i xs -> zip [i..] xs == zipFrom i xs
 -- > zipFrom False [1..3] == [(False,1),(True, 2)]
@@ -236,7 +273,7 @@ zipFrom :: Enum a => a -> [b] -> [(a, b)]
 zipFrom = zipWithFrom (,)
 
 -- | 'zipFrom' generalised to any combining operation.
---   Never truncates the output - raises an error if the enumeration runs out.
+--   Truncates the output if the enumeration runs out.
 --
 -- > \i xs -> zipWithFrom (,) i xs == zipFrom i xs
 zipWithFrom :: Enum a => (a -> b -> c) -> a -> [b] -> [c]
@@ -389,12 +426,26 @@ unescapeJSON [] = []
 
 
 -- | A version of 'group' where the equality is done on some extracted value.
-groupOn :: Eq b => (a -> b) -> [a] -> [[a]]
+--
+-- > groupOn abs [1,-1,2] == [[1,-1], [2]]
+groupOn :: Eq k => (a -> k) -> [a] -> [[a]]
 groupOn f = groupBy ((==) `on2` f)
     -- redefine on so we avoid duplicate computation for most values.
     where (.*.) `on2` f = \x -> let fx = f x in \y -> fx .*. f y
 
 
+-- | A version of 'groupOn' which pairs each group with its "key" - the
+--   extracted value used for equality testing.
+--
+-- > groupOnKey abs [1,-1,2] == [(1, [1,-1]), (2, [2])]
+groupOnKey :: Eq k => (a -> k) -> [a] -> [(k, [a])]
+groupOnKey _ []     = []
+groupOnKey f (x:xs) = (fx, x:yes) : groupOnKey f no
+    where
+        fx = f x
+        (yes, no) = span (\y -> fx == f y) xs
+
+
 -- | /DEPRECATED/ Use 'nubOrdOn', since this function is _O(n^2)_.
 --
 --   A version of 'nub' where the equality is done on some extracted value.
@@ -455,6 +506,32 @@ groupSortBy :: (a -> a -> Ordering) -> [a] -> [[a]]
 groupSortBy f = groupBy (\a b -> f a b == EQ) . sortBy f
 
 
+-- | A strict version of 'sum'.
+--   Unlike 'sum' this function is always strict in the `Num` argument,
+--   whereas the standard version is only strict if the optimiser kicks in.
+--
+-- > sum' [1, 2, 3] == 6
+sum' :: (Num a) => [a] -> a
+sum' = foldl' (+) 0
+
+-- | A strict version of 'sum', using a custom valuation function.
+--
+-- > sumOn' read ["1", "2", "3"] == 6
+sumOn' :: (Num b) => (a -> b) -> [a] -> b
+sumOn' f = foldl' (\acc x -> acc + f x) 0
+
+-- | A strict version of 'product'.
+--
+-- > product' [1, 2, 4] == 8
+product' :: (Num a) => [a] -> a
+product' = foldl' (*) 1
+
+-- | A strict version of 'product', using a custom valuation function.
+--
+-- > productOn' read ["1", "2", "4"] == 8
+productOn' :: (Num b) => (a -> b) -> [a] -> b
+productOn' f = foldl' (\acc x -> acc * f x) 1
+
 -- | Merge two lists which are assumed to be ordered.
 --
 -- > merge "ace" "bd" == "abcde"
@@ -816,3 +893,26 @@ zipWithLongest f [] [] = []
 zipWithLongest f (x:xs) (y:ys) = f (Just x) (Just y) : zipWithLongest f xs ys
 zipWithLongest f [] ys = map (f Nothing . Just) ys
 zipWithLongest f xs [] = map ((`f` Nothing) . Just) xs
+
+-- | Lazily compare the length of a 'Foldable' with a number.
+--
+-- > compareLength [1,2,3] 1 == GT
+-- > compareLength [1,2] 2 == EQ
+-- > \(xs :: [Int]) n -> compareLength xs n == compare (length xs) n
+-- > compareLength (1:2:3:undefined) 2 == GT
+compareLength :: (Ord b, Num b, Foldable f) => f a -> b -> Ordering
+compareLength = foldr (\_ acc n -> if n > 0 then acc (n - 1) else GT) (compare 0)
+
+-- | Lazily compare the length of two 'Foldable's.
+-- > comparingLength [1,2,3] [False] == GT
+-- > comparingLength [1,2] "ab" == EQ
+-- > \(xs :: [Int]) (ys :: [Int]) -> comparingLength xs ys == Data.Ord.comparing length xs ys
+-- > comparingLength [1,2] (1:2:3:undefined) == LT
+-- > comparingLength (1:2:3:undefined) [1,2] == GT
+comparingLength :: (Foldable f1, Foldable f2) => f1 a -> f2 b -> Ordering
+comparingLength x y = go (toList x) (toList y)
+  where
+    go [] [] = EQ
+    go [] (_:_) = LT
+    go (_:_) [] = GT
+    go (_:xs) (_:ys) = go xs ys
