@@ -14,7 +14,6 @@ import Test.QuickCheck
 import System.IO.Unsafe
 import Text.Show.Functions()
 
-import Control.Applicative as X
 import Control.Concurrent.Extra as X
 import Control.Exception.Extra as X
 import Control.Monad.Extra as X
@@ -84,8 +83,8 @@ instance Testable a => Testable (IO a) where
 -- So we print out full information on failure
 instance (Show a, Eq a) => Eq (IO a) where
     a == b = unsafePerformIO $ do
-        a <- try_ $ captureOutput a
-        b <- try_ $ captureOutput b
+        a <- captureOutput $ try_ a
+        b <- captureOutput $ try_ b
         if a == b then pure True else
             error $ show ("IO values not equal", a, b)
 
